from math import ceil, floor, trunc
from typing import Type, TypeAlias, Union

import pytest

from ethereum_types.bytes import Bytes4, Bytes8, Bytes64
from ethereum_types.numeric import (
    U32,
    U64,
    U256,
    FixedUnsigned,
    Uint,
    Unsigned,
    ulen,
)

FIXED_TYPES = (U256, U64, U32)

UNSIGNED_MARKS = (pytest.mark.unsigned,)
FIXED_MARKS = (pytest.mark.fixed,) + UNSIGNED_MARKS
UINT_MARKS = (pytest.mark.arbitrary,) + UNSIGNED_MARKS

FIXED = tuple(pytest.param(x, marks=FIXED_MARKS) for x in FIXED_TYPES)
UNSIGNED = (pytest.param(Uint, marks=UINT_MARKS),) + FIXED

FromBytesType: TypeAlias = Union[Type[Uint], Type[FixedUnsigned]]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_new(Class: Type[Unsigned]) -> None:
    value = Class(5)
    assert not isinstance(value, int)  # type: ignore[unreachable]
    assert isinstance(value, Class)
    assert isinstance(value, Unsigned)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_new_negative(Class: Type[Unsigned]) -> None:
    with pytest.raises(OverflowError):
        Class(-5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_new_float(Class: Type[Unsigned]) -> None:
    assert Class(0.1) == Class(0)
    assert Class(1.0) == Class(1)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_new_max_value(Class: Type[FixedUnsigned]) -> None:
    value = Class(Class.MAX_VALUE._number)
    assert isinstance(value, Class)
    assert isinstance(value, FixedUnsigned)
    assert value == Class.MAX_VALUE


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_new_too_large(Class: Type[FixedUnsigned]) -> None:
    with pytest.raises(OverflowError):
        Class(Class.MAX_VALUE._number + 1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_radd(Class: Type[Unsigned]) -> None:
    assert Class(9) == Class(4).__radd__(Class(5))


@pytest.mark.parametrize("Class", UNSIGNED)
def test_radd_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        4 + Class(5)  # type: ignore[operator]


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_radd_overflow(Class: Type[FixedUnsigned]) -> None:
    with pytest.raises(OverflowError):
        Class(Class.MAX_VALUE).__radd__(Class(5))
    with pytest.raises(OverflowError):
        Class(5).__radd__(Class(Class.MAX_VALUE))


@pytest.mark.parametrize("Class", UNSIGNED)
def test_radd_float(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        (1.0) + Class(5)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_add(Class: Type[Unsigned]) -> None:
    actual = Class(5) + Class(4)
    assert Class(9) == actual
    assert isinstance(actual, Class)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_add_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(5) + 4  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_add_float(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(5) + (1.0)  # type: ignore[operator]


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_add_overflow(Class: Type[FixedUnsigned]) -> None:
    with pytest.raises(OverflowError):
        Class(5) + Class.MAX_VALUE


@pytest.mark.parametrize("Class", UNSIGNED)
def test_iadd(Class: Type[Unsigned]) -> None:
    value = Class(5)
    value2 = value
    value += Class(4)
    assert isinstance(value, Class)
    assert value == Class(9)
    assert value2 == Class(5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_iadd_int(Class: Type[Unsigned]) -> None:
    value = Class(5)
    with pytest.raises(TypeError):
        value += 4  # type: ignore[arg-type]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_iadd_float(Class: Type[Unsigned]) -> None:
    value = Class(5)
    with pytest.raises(TypeError):
        value += 1.0  # type: ignore[arg-type]


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_iadd_overflow(Class: Type[FixedUnsigned]) -> None:
    value = Class(5)
    with pytest.raises(OverflowError):
        value += Class.MAX_VALUE
    assert value == Class(5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rsub(Class: Type[Unsigned]) -> None:
    actual = Class(5).__rsub__(Class(6))
    assert isinstance(actual, Class)
    assert Class(1) == actual


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rsub_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        6 - Class(5)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rsub_underflow(Class: Type[Unsigned]) -> None:
    with pytest.raises(OverflowError):
        Class(7).__rsub__(Class(6))


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rsub_float(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        (6.0) - Class(5)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_sub(Class: Type[Unsigned]) -> None:
    actual = Class(5) - Class(4)
    assert Class(1) == actual
    assert isinstance(actual, Class)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_sub_underflow(Class: Type[Unsigned]) -> None:
    with pytest.raises(OverflowError):
        Class(5) - Class(6)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_sub_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(5) - (-4)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_sub_float(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(5) - (1.0)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_isub(Class: Type[Unsigned]) -> None:
    value = Class(5)
    value2 = value
    value -= Class(4)
    assert isinstance(value, Class)
    assert value == Class(1)
    assert value2 == Class(5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_isub_underflow(Class: Type[Unsigned]) -> None:
    value = Class(5)
    with pytest.raises(OverflowError):
        value -= Class(6)
    assert value == Class(5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_isub_int(Class: Type[Unsigned]) -> None:
    value = Class(5)
    with pytest.raises(TypeError):
        value -= -4  # type: ignore[arg-type]
    assert value == Class(5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_isub_float(Class: Type[Unsigned]) -> None:
    value = Class(5)
    with pytest.raises(TypeError):
        value -= 1.0  # type: ignore[arg-type]
    assert value == Class(5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rmul(Class: Type[Unsigned]) -> None:
    actual = Class(4).__rmul__(Class(5))
    assert actual == Class(20)
    assert isinstance(actual, Class)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rmul_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        4 * Class(5)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rmul_float(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        (1.0) * Class(5)  # type: ignore[operator]


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_rmul_overflow(Class: Type[FixedUnsigned]) -> None:
    with pytest.raises(OverflowError):
        Class.MAX_VALUE.__rmul__(Class(5))


@pytest.mark.parametrize("Class", UNSIGNED)
def test_mul(Class: Type[Unsigned]) -> None:
    value = Class(5) * Class(4)
    assert isinstance(value, Class)
    assert value == Class(20)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_mul_int(Class: Type[Unsigned]) -> None:
    value = Class(5)
    with pytest.raises(TypeError):
        value * 4  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_mul_float(Class: Type[Unsigned]) -> None:
    value = Class(5)
    with pytest.raises(TypeError):
        value * (1.0)  # type: ignore[operator]


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_mul_overflow(Class: Type[FixedUnsigned]) -> None:
    four = Class(4)
    with pytest.raises(OverflowError):
        Class.MAX_VALUE * four


@pytest.mark.parametrize("Class", UNSIGNED)
def test_imul(Class: Type[Unsigned]) -> None:
    value = Class(5)
    value2 = value
    value *= Class(4)
    assert isinstance(value, Class)
    assert value == Class(20)
    assert value2 == Class(5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_imul_int(Class: Type[Unsigned]) -> None:
    value = Class(5)
    with pytest.raises(TypeError):
        value *= 4  # type: ignore[arg-type]
    assert value == Class(5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_imul_float(Class: Type[Unsigned]) -> None:
    value = Class(5)
    with pytest.raises(TypeError):
        value *= 1.0  # type: ignore[arg-type]
    assert value == Class(5)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_imul_overflow(Class: Type[FixedUnsigned]) -> None:
    value = Class(5)
    with pytest.raises(OverflowError):
        value *= Class.MAX_VALUE
    assert value == Class(5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_floordiv(Class: Type[Unsigned]) -> None:
    value = Class(5) // Class(2)
    assert isinstance(value, Class)
    assert value == Class(2)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_floordiv_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(5) // 2  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_floordiv_float(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(5) // 2.0  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rfloordiv(Class: Type[Unsigned]) -> None:
    value = Class(2).__rfloordiv__(Class(5))
    assert isinstance(value, Class)
    assert value == Class(2)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rfloordiv_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        (-2) // Class(5)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rfloordiv_float(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        5.0 // Class(2)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_ifloordiv(Class: Type[Unsigned]) -> None:
    value = Class(5)
    value2 = value
    value //= Class(2)
    assert isinstance(value, Class)
    assert value == Class(2)
    assert value2 == Class(5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_ifloordiv_int(Class: Type[Unsigned]) -> None:
    value = Class(5)
    with pytest.raises(TypeError):
        value //= -2  # type: ignore[arg-type]
    assert value == Class(5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rmod(Class: Type[Unsigned]) -> None:
    value = Class(2).__rmod__(Class(5))
    assert isinstance(value, Class)
    assert value == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rmod_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        (-4) % Class(5)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rmod_float(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        (6.0) % Class(5)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_mod(Class: Type[Unsigned]) -> None:
    value = Class(5) % Class(4)
    assert isinstance(value, Class)
    assert value == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_mod_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(5) % (-4)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_mod_float(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(5) % (1.0)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_imod(Class: Type[Unsigned]) -> None:
    value = Class(5)
    value2 = value
    value %= Class(4)
    assert isinstance(value, Class)
    assert value == Class(1)
    assert value2 == Class(5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_imod_int(Class: Type[Unsigned]) -> None:
    value = Class(5)
    with pytest.raises(TypeError):
        value %= -4  # type: ignore[arg-type]
    assert value == Class(5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_imod_float(Class: Type[Unsigned]) -> None:
    value = Class(5)
    with pytest.raises(TypeError):
        value %= 1.0  # type: ignore[arg-type]
    assert value == Class(5)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_divmod(Class: Type[Unsigned]) -> None:
    quotient, remainder = divmod(Class(5), Class(2))
    assert isinstance(quotient, Class)
    assert isinstance(remainder, Class)
    assert quotient == Class(2)
    assert remainder == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_divmod_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        divmod(Class(5), -2)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_divmod_float(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        divmod(Class(5), 2.0)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rdivmod(Class: Type[Unsigned]) -> None:
    quotient, remainder = Class(2).__rdivmod__(Class(5))
    assert quotient == Class(2)
    assert remainder == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rdivmod_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        divmod(5, Class(2))  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rdivmod_float(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        divmod(5.0, Class(2))  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_pow(Class: Type[Unsigned]) -> None:
    value = Class(3) ** Class(2)
    assert isinstance(value, Class)
    assert value == Class(9)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_pow_overflow(Class: Type[FixedUnsigned]) -> None:
    with pytest.raises(OverflowError):
        Class(340282366920938463463374607431768211456) ** Class(3)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_pow_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(3) ** -2  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_pow_modulo(Class: Type[Unsigned]) -> None:
    value = pow(Class(4), Class(2), Class(3))
    assert isinstance(value, Class)
    assert value == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_pow_modulo_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        pow(Class(4), Class(2), -3)  # type: ignore[misc]

    with pytest.raises(TypeError):
        pow(Class(4), 2, Class(3))  # type: ignore[misc]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rpow(Class: Type[Unsigned]) -> None:
    value = Class(2).__rpow__(Class(3))
    assert isinstance(value, Class)
    assert value == Class(9)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rpow_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        (-3) ** Class(2)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rpow_modulo(Class: Type[Unsigned]) -> None:
    value = Class.__rpow__(Class(2), Class(4), Class(3))
    assert isinstance(value, Class)
    assert value == Class(1)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_rpow_overflow(Class: Type[FixedUnsigned]) -> None:
    with pytest.raises(OverflowError):
        Class.MAX_VALUE ** Class(2)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rpow_modulo_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class.__rpow__(Class(2), Class(4), -3)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_ipow(Class: Type[Unsigned]) -> None:
    value = Class(3)
    value2 = value
    value **= Class(2)
    assert isinstance(value, Class)
    assert value == Class(9)
    assert value2 == Class(3)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_ipow_int(Class: Type[Unsigned]) -> None:
    value = Class(3)
    with pytest.raises(TypeError):
        value **= -2  # type: ignore[arg-type]


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_ipow_overflow(Class: Type[FixedUnsigned]) -> None:
    value = Class.MAX_VALUE
    with pytest.raises(OverflowError):
        value **= Class(3)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_ipow_modulo(Class: Type[Unsigned]) -> None:
    value = Class(4).__ipow__(Class(2), Class(3))
    assert isinstance(value, Class)
    assert value == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_ipow_modulo_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(4).__ipow__(Class(2), -3)  # type: ignore[arg-type]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bit_length(Class: Type[Unsigned]) -> None:
    assert Class(3).bit_length() == Class(2)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_bytes_zero(Class: Type[Unsigned]) -> None:
    encoded = Class(0).to_bytes()
    assert encoded == bytes([0])


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_bytes(Class: Type[Unsigned]) -> None:
    encoded = Class(1).to_bytes(length=Uint(5))
    assert encoded == bytes([0, 0, 0, 0, 1])


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_be_bytes_zero(Class: Type[Unsigned]) -> None:
    encoded = Class(0).to_be_bytes()
    assert encoded == bytes([])


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_be_bytes_one(Class: Type[Unsigned]) -> None:
    encoded = Class(1).to_be_bytes()
    assert encoded == bytes([1])


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_be_bytes_is_big_endian(Class: Type[Unsigned]) -> None:
    encoded = Class(0xABCD).to_be_bytes()
    assert encoded == bytes([0xAB, 0xCD])


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_be_bytes64_zero(Class: Type[Unsigned]) -> None:
    actual = Class(0).to_be_bytes64()
    expected = Bytes64([0] * 64)
    assert isinstance(actual, Bytes64)
    assert actual == expected


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_be_bytes64_one(Class: Type[Unsigned]) -> None:
    actual = Class(1).to_be_bytes64()
    expected = Bytes64([0] * 63 + [1])
    assert isinstance(actual, Bytes64)
    assert actual == expected


@pytest.mark.unsigned
@pytest.mark.arbitrary
def test_uint_to_be_bytes64_max_value() -> None:
    actual = Uint(2**512 - 1).to_be_bytes64()
    expected = Bytes64([0xFF] * 64)
    assert isinstance(actual, Bytes64)
    assert actual == expected


@pytest.mark.unsigned
@pytest.mark.arbitrary
def test_uint_to_be_bytes64_too_big() -> None:
    value = Uint(2**512)

    with pytest.raises(OverflowError):
        value.to_be_bytes64()


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_be_bytes32_zero(Class: Type[Unsigned]) -> None:
    encoded = Class(0).to_be_bytes32()
    assert encoded == bytes([0] * 32)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_be_bytes32_one(Class: Type[Unsigned]) -> None:
    encoded = Class(1).to_be_bytes32()
    assert encoded == bytes([0] * 31 + [1])


@pytest.mark.unsigned
@pytest.mark.arbitrary
def test_uint_to_be_bytes32_max_value() -> None:
    encoded = Uint(2**256 - 1).to_be_bytes32()
    assert encoded == bytes([0xFF] * 32)


@pytest.mark.unsigned
@pytest.mark.fixed
def test_u256_to_be_bytes32_max_value() -> None:
    encoded = U256(2**256 - 1).to_be_bytes32()
    assert encoded == bytes([0xFF] * 32)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_from_be_bytes_empty(Class: FromBytesType) -> None:
    value = Class.from_be_bytes(b"")
    assert value == Class(0)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_from_be_bytes_one(Class: FromBytesType) -> None:
    value = Class.from_be_bytes(bytes([1]))
    assert value == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_from_be_bytes_is_big_endian(Class: FromBytesType) -> None:
    value = Class.from_be_bytes(bytes([0xAB, 0xCD]))
    assert value == Class(0xABCD)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_from_le_bytes_empty(Class: FromBytesType) -> None:
    value = Class.from_le_bytes(b"")
    assert value == Class(0)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_from_le_bytes_one(Class: FromBytesType) -> None:
    value = Class.from_le_bytes(bytes([1]))
    assert value == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_from_le_bytes_is_little_endian(Class: FromBytesType) -> None:
    value = Class.from_le_bytes(bytes([0xAB, 0xCD]))
    assert value == Class(0xCDAB)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_abs(Class: Type[Unsigned]) -> None:
    assert abs(Class(1)) == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_abs_same(Class: Type[Unsigned]) -> None:
    a = Class(1)
    b = abs(a)
    a += Class(1)
    assert a == Class(2)
    assert b == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_trunc(Class: Type[Unsigned]) -> None:
    assert trunc(Class(1)) == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_trunc_same(Class: Type[Unsigned]) -> None:
    a = Class(1)
    b = trunc(a)
    a += Class(1)
    assert a == Class(2)
    assert b == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_round(Class: Type[Unsigned]) -> None:
    assert round(Class(1)) == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_round_same(Class: Type[Unsigned]) -> None:
    a = Class(1)
    b = round(a)
    a += Class(1)
    assert a == Class(2)
    assert b == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_truediv_int(Class: Type[Unsigned]) -> None:
    assert Class(1).__truediv__(2) is NotImplemented  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rtruediv_int(Class: Type[Unsigned]) -> None:
    assert Class(1).__rtruediv__(2) is NotImplemented  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_truediv(Class: Type[Unsigned]) -> None:
    expected = (1).__truediv__(2)
    actual = Class(1).__truediv__(Class(2))
    assert expected == actual


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rtruediv(Class: Type[Unsigned]) -> None:
    expected = (1).__rtruediv__(2)
    actual = Class(1).__rtruediv__(Class(2))
    assert expected == actual


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_wrapping_add(Class: Type[FixedUnsigned]) -> None:
    value = Class(5).wrapping_add(Class(4))
    assert isinstance(value, Class)
    assert value == 9


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_wrapping_add_overflow(Class: Type[FixedUnsigned]) -> None:
    value = Class(5).wrapping_add(Class.MAX_VALUE)
    assert isinstance(value, Class)
    assert value == 4


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_wrapping_add_int(Class: Type[FixedUnsigned]) -> None:
    with pytest.raises(TypeError):
        Class(5).wrapping_add(-4)  # type: ignore[arg-type]


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_wrapping_sub(Class: Type[FixedUnsigned]) -> None:
    value = Class(5).wrapping_sub(Class(4))
    assert isinstance(value, Class)
    assert value == 1


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_wrapping_sub_underflow(Class: Type[FixedUnsigned]) -> None:
    value = Class(5).wrapping_sub(Class(6))
    assert isinstance(value, Class)
    assert value == Class.MAX_VALUE


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_wrapping_sub_int(Class: Type[FixedUnsigned]) -> None:
    with pytest.raises(TypeError):
        Class(5).wrapping_sub(-4)  # type: ignore[arg-type]


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_wrapping_mul(Class: Type[FixedUnsigned]) -> None:
    value = Class(5).wrapping_mul(Class(4))
    assert isinstance(value, Class)
    assert value == Class(20)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_wrapping_mul_overflow(Class: Type[FixedUnsigned]) -> None:
    value = Class.MAX_VALUE.wrapping_mul(Class(4))
    expected = Class.MAX_VALUE - Class(3)
    assert isinstance(value, Class)
    assert value == expected


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_wrapping_mul_int(Class: Type[FixedUnsigned]) -> None:
    with pytest.raises(TypeError):
        Class(5).wrapping_mul(-4)  # type: ignore[arg-type]


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_wrapping_pow(Class: Type[FixedUnsigned]) -> None:
    value = Class(3).wrapping_pow(Class(2))
    assert isinstance(value, Class)
    assert value == Class(9)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_wrapping_pow_overflow(Class: Type[FixedUnsigned]) -> None:
    base = (Class.MAX_VALUE // Class(3)) - Class(5)
    actual = base.wrapping_pow(Class(3))
    expected = pow(int(base), 3, int(Class.MAX_VALUE) + 1)
    assert isinstance(actual, Class)
    assert expected == actual


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_wrapping_pow_int(Class: Type[FixedUnsigned]) -> None:
    with pytest.raises(TypeError):
        Class(3).wrapping_pow(-2)  # type: ignore[arg-type]


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_wrapping_pow_modulo(Class: Type[FixedUnsigned]) -> None:
    value = Class(4).wrapping_pow(Class(2), Class(3))
    assert isinstance(value, Class)
    assert value == Class(1)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_wrapping_pow_modulo_int(Class: Type[FixedUnsigned]) -> None:
    with pytest.raises(TypeError):
        Class(4).wrapping_pow(Class(2), 2)  # type: ignore[arg-type]


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_from_be_bytes_too_large(Class: Type[FixedUnsigned]) -> None:
    bits = Class.MAX_VALUE._number.bit_length()
    byte_count = (bits + 7) // 8
    byte_count += 1

    with pytest.raises(ValueError):
        Class.from_be_bytes(bytes([0xFF] * byte_count))


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_from_le_bytes_too_large(Class: Type[FixedUnsigned]) -> None:
    bits = Class.MAX_VALUE._number.bit_length()
    byte_count = (bits + 7) // 8
    byte_count += 1

    with pytest.raises(ValueError):
        Class.from_le_bytes(bytes([0xFF] * byte_count))


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_invert(Class: Type[FixedUnsigned]) -> None:
    assert ~Class(0) == Class.MAX_VALUE
    assert ~Class(10) == Class.MAX_VALUE - Class(10)
    assert ~Class.MAX_VALUE == 0


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_rshift(Class: Type[FixedUnsigned]) -> None:
    bits = Class(Class.MAX_VALUE.bit_length())

    assert Class.MAX_VALUE >> (bits - Class(1)) == Class(1)
    assert Class.MAX_VALUE >> Class(bits) == Class(0)
    assert Class.MAX_VALUE >> Class(bits + Class(1)) == Class(0)
    assert Class(0) >> Class(20) == Class(0)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_le_bytes_zero(Class: Type[Unsigned]) -> None:
    encoded = Class(0).to_le_bytes()
    assert encoded == bytes([])


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_le_bytes_one(Class: Type[Unsigned]) -> None:
    encoded = Class(1).to_le_bytes()
    assert encoded == bytes([1])


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_le_bytes_is_little_endian(Class: Type[Unsigned]) -> None:
    encoded = Class(0xABCD).to_le_bytes()
    assert encoded == bytes([0xCD, 0xAB])


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_le_bytes64_zero(Class: Type[Unsigned]) -> None:
    actual = Class(0).to_le_bytes64()
    expected = Bytes64([0] * 64)
    assert isinstance(actual, Bytes64)
    assert actual == expected


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_le_bytes64_one(Class: Type[Unsigned]) -> None:
    actual = Class(1).to_le_bytes64()
    expected = Bytes64([1] + [0] * 63)
    assert isinstance(actual, Bytes64)
    assert actual == expected


@pytest.mark.unsigned
@pytest.mark.arbitrary
def test_uint_to_le_bytes64_max_value() -> None:
    actual = Uint(2**512 - 1).to_le_bytes64()
    expected = Bytes64([0xFF] * 64)
    assert isinstance(actual, Bytes64)
    assert actual == expected


@pytest.mark.unsigned
@pytest.mark.arbitrary
def test_uint_to_le_bytes64_too_big() -> None:
    value = Uint(2**512)

    with pytest.raises(OverflowError):
        value.to_le_bytes64()


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_le_bytes32_zero(Class: Type[Unsigned]) -> None:
    encoded = Class(0).to_le_bytes32()
    assert encoded == bytes([0] * 32)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_le_bytes32_one(Class: Type[Unsigned]) -> None:
    encoded = Class(1).to_le_bytes32()
    assert encoded == bytes([1] + ([0] * 31))


@pytest.mark.unsigned
@pytest.mark.arbitrary
def test_uint_to_le_bytes32_max_value() -> None:
    encoded = Uint(2**256 - 1).to_le_bytes32()
    assert encoded == bytes([0xFF] * 32)


@pytest.mark.unsigned
@pytest.mark.fixed
def test_u256_to_le_bytes32_max_value() -> None:
    encoded = U256(2**256 - 1).to_le_bytes32()
    assert encoded == bytes([0xFF] * 32)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_neg(Class: Type[Unsigned]) -> None:
    result = -Class(1)
    assert not isinstance(result, Unsigned)  # type: ignore[unreachable]
    assert result == -1


@pytest.mark.parametrize("Class", UNSIGNED)
def test_pos(Class: Type[Unsigned]) -> None:
    assert Class(1) == +Class(1)


@pytest.mark.unsigned
@pytest.mark.fixed
def test_uint_invert() -> None:
    with pytest.raises(NotImplementedError):
        ~Uint(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_int(Class: Type[Unsigned]) -> None:
    assert int(Class(0xF000)) == Class(0xF000)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_floor_same(Class: Type[Unsigned]) -> None:
    a = Class(1)
    b = floor(a)
    a += Class(1)
    assert a == Class(2)
    assert b == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_floor(Class: Type[Unsigned]) -> None:
    assert floor(Class(1)) == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_ceil_same(Class: Type[Unsigned]) -> None:
    a = Class(1)
    b = ceil(a)
    a += Class(1)
    assert a == Class(2)
    assert b == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_ceil(Class: Type[Unsigned]) -> None:
    assert ceil(Class(1)) == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_ne(Class: Type[Unsigned]) -> None:
    assert Class(1) != Class(2)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_ne_not(Class: Type[Unsigned]) -> None:
    assert not (Class(1) != Class(1))


@pytest.mark.parametrize("Class", UNSIGNED)
def test_le(Class: Type[Unsigned]) -> None:
    assert Class(1) <= Class(1)
    assert Class(0) <= Class(1)
    assert not (Class(2) <= Class(1))


@pytest.mark.parametrize("Class", UNSIGNED)
def test_le_different_types(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(1) <= 1  # type: ignore[operator] # noqa: B015


@pytest.mark.parametrize("Class", UNSIGNED)
def test_ge(Class: Type[Unsigned]) -> None:
    assert Class(1) >= Class(1)
    assert Class(2) >= Class(1)
    assert not (Class(1) >= Class(2))


@pytest.mark.parametrize("Class", UNSIGNED)
def test_ge_different_types(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(1) >= 1  # type: ignore[operator] # noqa: B015


@pytest.mark.parametrize("Class", UNSIGNED)
def test_lt(Class: Type[Unsigned]) -> None:
    assert not (Class(1) < Class(1))
    assert Class(0) < Class(1)
    assert not (Class(2) < Class(1))


@pytest.mark.parametrize("Class", UNSIGNED)
def test_lt_different_types(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(1) < 1  # type: ignore[operator] # noqa: B015
    with pytest.raises(TypeError):
        1 < Class(1)  # type: ignore[operator] # noqa: B015


@pytest.mark.parametrize("Class", UNSIGNED)
def test_gt(Class: Type[Unsigned]) -> None:
    assert not (Class(1) > Class(1))
    assert Class(2) > Class(1)
    assert not (Class(1) > Class(2))


@pytest.mark.parametrize("Class", UNSIGNED)
def test_gt_different_types(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(1) > 1  # type: ignore[operator] # noqa: B015
    with pytest.raises(TypeError):
        1 > Class(1)  # type: ignore[operator] # noqa: B015


@pytest.mark.parametrize("Class", UNSIGNED)
def test_lshift_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(3) << 4  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_lshift(Class: Type[Unsigned]) -> None:
    expected = Class(3 << 4)
    actual = Class(3) << Class(4)
    assert expected == actual


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rlshift_int(Class: Type[Unsigned]) -> None:
    assert Class(3).__rlshift__(4) is NotImplemented  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rlshift(Class: Type[Unsigned]) -> None:
    expected = Class((3).__rlshift__(4))
    actual = Class(3).__rlshift__(Class(4))
    assert expected == actual


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rshift(Class: Type[Unsigned]) -> None:
    expected = Class(3 >> 4)
    actual = Class(3) >> Class(4)
    assert expected == actual


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rshift_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(3) >> 4  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rrshift(Class: Type[Unsigned]) -> None:
    expected = Class((3).__rrshift__(4))
    actual = Class(3).__rrshift__(Class(4))
    assert expected == actual


@pytest.mark.parametrize("Class", UNSIGNED)
def test_rrshift_int(Class: Type[Unsigned]) -> None:
    assert Class(3).__rrshift__(4) is NotImplemented  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_eq_str(Class: Type[Unsigned]) -> None:
    assert Class(1).__eq__("hello") is NotImplemented


@pytest.mark.parametrize("Class", UNSIGNED)
def test_eq_float(Class: Type[Unsigned]) -> None:
    assert Class(1) == 1.0
    assert 1.0 == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_eq_int(Class: Type[Unsigned]) -> None:
    assert Class(1) == 1
    assert 1 == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_eq(Class: Type[Unsigned]) -> None:
    assert Class(1) == Class(1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_eq_not_float(Class: Type[Unsigned]) -> None:
    assert not (Class(1) == 1.1)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_eq_not(Class: Type[Unsigned]) -> None:
    assert not (Class(1) == Class(2))


@pytest.mark.parametrize("Class", UNSIGNED)
def test_hash(Class: Type[Unsigned]) -> None:
    assert hash(Class(1)) == hash(Class(1))
    assert hash(Class(1)) != hash(Class(2))


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bitwise_rand(Class: Type[Unsigned]) -> None:
    assert Class(0).__rand__(Class(0)) == Class(0)
    assert Class(2**32 - 1).__rand__(Class(2**32 - 1)) == Class(2**32 - 1)
    assert Class(2**32 - 1).__rand__(Class(0)) == Class(0)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_bitwise_rand(Class: Type[FixedUnsigned]) -> None:
    assert Class.MAX_VALUE.__rand__(Class.MAX_VALUE) == Class(Class.MAX_VALUE)
    assert Class.MAX_VALUE.__rand__(Class(0)) == Class(0)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bitwise_rand_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        (2**256) & Class(0)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bitwise_and(Class: Type[Unsigned]) -> None:
    assert Class(0) & Class(0) == Class(0)
    assert Class(2**32 - 1) & Class(2**32 - 1) == Class(2**32 - 1)
    assert Class(2**32 - 1) & Class(0) == Class(0)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_bitwise_and(Class: Type[FixedUnsigned]) -> None:
    assert Class.MAX_VALUE & Class.MAX_VALUE == Class.MAX_VALUE
    assert Class.MAX_VALUE.__rand__(Class(0)) == Class(0)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bitwise_and_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(0) & (2**256)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bitwise_ror(Class: Type[Unsigned]) -> None:
    assert Class(0).__ror__(Class(0)) == Class(0)
    assert Class(2**32 - 1).__ror__(Class(0)) == Class(2**32 - 1)
    assert Class(2**32 - 1).__ror__(Class(2**32 - 1)) == Class(2**32 - 1)
    assert Class(2**32 - 1).__ror__(Class(17)) == Class(2**32 - 1)
    assert Class(17).__ror__(Class(18)) == Class(19)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_bitwise_ror(Class: Type[FixedUnsigned]) -> None:
    assert Class.MAX_VALUE.__ror__(Class(0)) == Class.MAX_VALUE
    assert Class.MAX_VALUE.__ror__(Class.MAX_VALUE) == Class.MAX_VALUE
    assert Class.MAX_VALUE.__ror__(Class(17)) == Class.MAX_VALUE


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bitwise_ror_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        (2**256) | Class(0)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bitwise_or(Class: Type[Unsigned]) -> None:
    assert Class(0) | Class(0) == Class(0)
    assert Class(2**32 - 1) | Class(0) == Class(2**32 - 1)
    assert Class(2**32 - 1) | Class(2**32 - 1) == Class(2**32 - 1)
    assert Class(2**32 - 1) | Class(17) == Class(2**32 - 1)
    assert Class(17) | Class(18) == Class(19)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_bitwise_or(Class: Type[FixedUnsigned]) -> None:
    assert Class.MAX_VALUE | Class(0) == Class.MAX_VALUE
    assert Class.MAX_VALUE | Class.MAX_VALUE == Class.MAX_VALUE
    assert Class.MAX_VALUE | Class(17) == Class.MAX_VALUE


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bitwise_or_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(0) | (2**256)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bitwise_xor(Class: Type[Unsigned]) -> None:
    assert Class(0) ^ Class(0) == Class(0)
    assert Class(2**32 - 1) ^ Class(0) == Class(2**32 - 1)
    assert Class(2**32 - 1) ^ Class(2**32 - 1) == Class(0)
    assert Class(2**32 - 1) ^ Class(17) == Class(2**32 - 1) - Class(17)
    assert Class(17) ^ Class(18) == Class(3)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_bitwise_xor(Class: Type[FixedUnsigned]) -> None:
    assert Class.MAX_VALUE ^ Class(0) == Class.MAX_VALUE
    assert Class.MAX_VALUE ^ Class.MAX_VALUE == Class(0)
    assert Class.MAX_VALUE ^ Class(17) == Class.MAX_VALUE - Class(17)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bitwise_xor_int(Class: Type[Unsigned]) -> None:
    with pytest.raises(TypeError):
        Class(0) ^ (2**256)  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bitwise_rxor(Class: Type[Unsigned]) -> None:
    assert Class(0).__rxor__(Class(0)) == Class(0)
    assert Class(2**32 - 1).__rxor__(Class(0)) == Class(2**32 - 1)
    assert Class(2**32 - 1).__rxor__(Class(2**32 - 1)) == Class(0)
    assert Class(2**32 - 1).__rxor__(Class(17)) == Class(2**32 - 1) - Class(17)
    assert Class(17).__rxor__(Class(18)) == Class(3)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_bitwise_rxor(Class: Type[FixedUnsigned]) -> None:
    assert Class.MAX_VALUE.__rxor__(Class(0)) == Class.MAX_VALUE
    assert Class.MAX_VALUE.__rxor__(Class.MAX_VALUE) == Class(0)
    assert Class.MAX_VALUE.__rxor__(Class(17)) == Class.MAX_VALUE - Class(17)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bitwise_rxor_int(Class: Type[Unsigned]) -> None:
    assert Uint(0).__rxor__(2**256) is NotImplemented  # type: ignore[operator]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bitwise_ixor(Class: Type[Unsigned]) -> None:
    value = Class(1)
    value2 = value
    value ^= Class(1)
    assert value == Class(0)
    assert value2 == Class(1)

    assert Class(0).__ixor__(Class(0)) == Class(0)
    assert Class(2**32 - 1).__ixor__(Class(0)) == Class(2**32 - 1)
    assert Class(2**32 - 1).__ixor__(Class(2**32 - 1)) == Class(0)
    assert Class(2**32 - 1).__ixor__(Class(17)) == Class(2**32 - 1) - Class(17)
    assert Class(17).__ixor__(Class(18)) == Class(3)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_bitwise_ixor(Class: Type[FixedUnsigned]) -> None:
    assert Class.MAX_VALUE.__ixor__(Class(0)) == Class.MAX_VALUE
    assert Class.MAX_VALUE.__ixor__(Class.MAX_VALUE) == Class(0)
    assert Class.MAX_VALUE.__ixor__(Class(17)) == Class.MAX_VALUE - Class(17)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_bitwise_ixor_int(Class: Type[Unsigned]) -> None:
    assert Class(0).__ixor__(2**256) is NotImplemented  # type: ignore[arg-type]


@pytest.mark.parametrize("Class", UNSIGNED)
def test_repr(Class: Type[Unsigned]) -> None:
    assert repr(Class(1)) == f"{Class.__name__}(1)"


@pytest.mark.parametrize("Class", UNSIGNED)
def test_str(Class: Type[Unsigned]) -> None:
    assert str(Class(1)) == "1"


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_le_bytes4_zero(Class: Type[Unsigned]) -> None:
    actual = Class(0).to_le_bytes4()
    assert isinstance(actual, Bytes4)
    assert actual == bytes([0] * 4)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_le_bytes4(Class: Type[Unsigned]) -> None:
    actual = Class(0x89ABCDEF).to_le_bytes4()
    assert isinstance(actual, Bytes4)
    assert actual == bytes([0xEF, 0xCD, 0xAB, 0x89])


@pytest.mark.unsigned
@pytest.mark.arbitrary
def test_uint_to_le_bytes4_overflow() -> None:
    value = Uint(0x0189ABCDEF)
    with pytest.raises(OverflowError):
        value.to_le_bytes4()


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_be_bytes4_zero(Class: Type[Unsigned]) -> None:
    actual = Class(0).to_be_bytes4()
    assert isinstance(actual, Bytes4)
    assert actual == bytes([0] * 4)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_be_bytes4(Class: Type[Unsigned]) -> None:
    actual = Class(0xEFCDAB89).to_be_bytes4()
    assert isinstance(actual, Bytes4)
    assert actual == bytes([0xEF, 0xCD, 0xAB, 0x89])


@pytest.mark.unsigned
@pytest.mark.arbitrary
def test_uint_to_be_bytes4_overflow() -> None:
    value = Uint(0x0189ABCDEF)
    with pytest.raises(OverflowError):
        value.to_be_bytes4()


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_le_bytes8_zero(Class: Type[Unsigned]) -> None:
    actual = Class(0).to_le_bytes8()
    assert isinstance(actual, Bytes8)
    assert actual == bytes([0] * 8)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_le_bytes8(Class: Type[Unsigned]) -> None:
    actual = Class(0x0000000089ABCDEF).to_le_bytes8()
    assert isinstance(actual, Bytes8)
    assert actual == bytes([0xEF, 0xCD, 0xAB, 0x89, 0x00, 0x00, 0x00, 0x00])


@pytest.mark.unsigned
@pytest.mark.arbitrary
def test_uint_to_le_bytes8() -> None:
    actual = Uint(0x0123456789ABCDEF).to_le_bytes8()
    assert isinstance(actual, Bytes8)
    assert actual == bytes([0xEF, 0xCD, 0xAB, 0x89, 0x67, 0x45, 0x23, 0x01])


@pytest.mark.unsigned
@pytest.mark.arbitrary
def test_uint_to_le_bytes8_overflow() -> None:
    value = Uint(0x010123456789ABCDEF)
    with pytest.raises(OverflowError):
        value.to_le_bytes8()


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_be_bytes8_zero(Class: Type[Unsigned]) -> None:
    actual = Class(0).to_be_bytes8()
    assert isinstance(actual, Bytes8)
    assert actual == bytes([0] * 8)


@pytest.mark.parametrize("Class", UNSIGNED)
def test_to_be_bytes8(Class: Type[Unsigned]) -> None:
    actual = Class(0x00000000EFCDAB89).to_be_bytes8()
    assert isinstance(actual, Bytes8)
    assert actual == bytes([0x00, 0x00, 0x00, 0x00, 0xEF, 0xCD, 0xAB, 0x89])


@pytest.mark.unsigned
@pytest.mark.arbitrary
def test_uint_to_be_bytes8() -> None:
    actual = Uint(0xEFCDAB8967452301).to_be_bytes8()
    assert isinstance(actual, Bytes8)
    assert actual == bytes([0xEF, 0xCD, 0xAB, 0x89, 0x67, 0x45, 0x23, 0x01])


@pytest.mark.unsigned
@pytest.mark.arbitrary
def test_uint_to_be_bytes8_overflow() -> None:
    value = Uint(0x010123456789ABCDEF)
    with pytest.raises(OverflowError):
        value.to_be_bytes8()


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_to_signed(Class: Type[FixedUnsigned]) -> None:
    value = Class(1234567890).to_signed()
    assert isinstance(value, int)
    assert value == 1234567890


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_to_signed_max_value(Class: Type[FixedUnsigned]) -> None:
    value = Class.MAX_VALUE.to_signed()
    assert isinstance(value, int)
    assert value == -1


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_from_signed_to_signed(Class: Type[FixedUnsigned]) -> None:
    unsigned = Class.from_signed(-123456)
    assert isinstance(unsigned, Class)

    signed = unsigned.to_signed()
    assert isinstance(signed, int)
    assert signed == -123456


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_from_signed(Class: Type[FixedUnsigned]) -> None:
    unsigned = Class.from_signed(1)
    assert isinstance(unsigned, Class)
    assert unsigned == Class(1)


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_from_signed_negative(Class: Type[FixedUnsigned]) -> None:
    unsigned = Class.from_signed(-1)
    assert isinstance(unsigned, Class)
    assert unsigned == Class.MAX_VALUE


@pytest.mark.parametrize("Class", FIXED)
def test_fixed_from_signed_overflow(Class: Type[FixedUnsigned]) -> None:
    too_positive = int(Class.MAX_VALUE) // 2 + 1
    with pytest.raises(OverflowError):
        Class.from_signed(too_positive)

    too_negative = (-Class.MAX_VALUE // 2) - 1
    with pytest.raises(OverflowError):
        Class.from_signed(too_negative)


def test_unsigned_in_range() -> None:
    number = Uint(0)
    with pytest.raises(NotImplementedError):
        Unsigned._in_range(number, 0)


def test_ulen() -> None:
    actual = ulen([1, 2, 3])
    assert isinstance(actual, Uint)
    assert actual == Uint(3)



"""
Numeric types (mostly integers.)
"""

from abc import abstractmethod
from numbers import Integral
from typing import (
    ClassVar,
    Final,
    Literal,
    Optional,
    Sized,
    SupportsInt,
    Tuple,
    Type,
    TypeVar,
)

from typing_extensions import Self, override

from .bytes import Bytes, Bytes4, Bytes8, Bytes32, Bytes64


class Unsigned(Integral):
    """
    Base of integer types.
    """

    __slots__ = ("_number",)
    _number: Final[int]

    def __init__(self, value: SupportsInt) -> None:
        int_value = int(value)
        if not self._in_range(int_value):
            raise OverflowError()
        self._number = int_value

    @abstractmethod
    def _in_range(self, value: int) -> bool:
        raise NotImplementedError

    @override
    def __abs__(self) -> Self:
        return type(self)(self)

    @override
    def __radd__(self, left: Self) -> Self:
        return self.__add__(left)

    @override
    def __add__(self, right: Self) -> Self:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented
        return Class(self._number + right._number)

    def __iadd__(self, right: Self) -> Self:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented
        return Class(self._number + right._number)

    @override
    def __sub__(self, right: Self) -> Self:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented

        if self._number < right._number:
            raise OverflowError()

        return Class(self._number - right._number)

    @override
    def __rsub__(self, left: Self) -> Self:
        Class = type(self)
        if not isinstance(left, Class):
            return NotImplemented

        if self._number > left._number:
            raise OverflowError()

        return Class(left._number - self._number)

    def __isub__(self, right: Self) -> Self:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented
        if right._number > self._number:
            raise OverflowError()
        return Class(self._number - right._number)

    @override
    def __mul__(self, right: Self) -> Self:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented
        return Class(self._number * right._number)

    @override
    def __rmul__(self, left: Self) -> Self:
        return self.__mul__(left)

    def __imul__(self, right: Self) -> Self:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented
        return Class(self._number * right._number)

    @override
    def __truediv__(self, other: Self) -> float:
        Class = type(self)
        if not isinstance(other, Class):
            return NotImplemented
        return self._number.__truediv__(other._number)

    @override
    def __rtruediv__(self, other: Self) -> float:
        Class = type(self)
        if not isinstance(other, Class):
            return NotImplemented
        return self._number.__rtruediv__(other._number)

    @override
    def __floordiv__(self, right: Self) -> Self:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented

        return Class(self._number.__floordiv__(right._number))

    @override
    def __rfloordiv__(self, left: Self) -> Self:
        Class = type(self)
        if not isinstance(left, Class):
            return NotImplemented

        return Class(self._number.__rfloordiv__(left._number))

    def __ifloordiv__(self, right: Self) -> Self:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented
        return Class(self._number // right._number)

    @override
    def __mod__(self, right: Self) -> Self:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented

        return Class(self._number % right._number)

    @override
    def __rmod__(self, left: Self) -> Self:
        Class = type(self)
        if not isinstance(left, Class):
            return NotImplemented

        return Class(self._number.__rmod__(left._number))

    def __imod__(self, right: Self) -> Self:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented
        return Class(self._number % right._number)

    @override
    def __divmod__(self, right: Self) -> Tuple[Self, Self]:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented

        result = self._number.__divmod__(right._number)
        return (
            Class(result[0]),
            Class(result[1]),
        )

    @override
    def __rdivmod__(self, left: Self) -> Tuple[Self, Self]:
        Class = type(self)
        if not isinstance(left, Class):
            return NotImplemented

        result = self._number.__rdivmod__(left._number)
        return (
            Class(result[0]),
            Class(result[1]),
        )

    @override
    def __pow__(self, right: Self, modulo: Optional[Self] = None) -> Self:
        Class = type(self)
        modulo_int = None
        if modulo is not None:
            if not isinstance(modulo, Class):
                return NotImplemented
            modulo_int = modulo._number

        if not isinstance(right, Class):
            return NotImplemented

        return Class(self._number.__pow__(right._number, modulo_int))

    @override
    def __rpow__(self, left: Self, modulo: Optional[Self] = None) -> Self:
        Class = type(self)
        modulo_int = None
        if modulo is not None:
            if not isinstance(modulo, Class):
                raise TypeError()
            modulo_int = modulo._number

        if not isinstance(left, Class):
            return NotImplemented

        return Class(self._number.__rpow__(left._number, modulo_int))

    def __ipow__(self, right: Self, modulo: Optional[Self] = None) -> Self:
        Class = type(self)
        modulo_int = None
        if modulo is not None:
            if not isinstance(modulo, Class):
                raise TypeError()
            modulo_int = modulo._number

        if not isinstance(right, Class):
            return NotImplemented

        return Class(self._number.__pow__(right._number, modulo_int))

    @override
    def __xor__(self, right: Self) -> Self:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented

        return Class(self._number.__xor__(right._number))

    @override
    def __rxor__(self, left: Self) -> Self:
        Class = type(self)
        if not isinstance(left, Class):
            return NotImplemented

        return Class(self._number.__rxor__(left._number))

    def __ixor__(self, right: Self) -> Self:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented

        return Class(self._number.__xor__(right._number))

    @override
    def __and__(self, other: Self) -> Self:
        Class = type(self)
        if not isinstance(other, Class):
            return NotImplemented

        return Class(self._number.__and__(other._number))

    @override
    def __rand__(self, other: Self) -> Self:
        Class = type(self)
        if not isinstance(other, Class):
            return NotImplemented

        return Class(self._number.__rand__(other._number))

    @override
    def __or__(self, other: Self) -> Self:
        Class = type(self)
        if not isinstance(other, Class):
            return NotImplemented

        return Class(self._number.__or__(other._number))

    @override
    def __ror__(self, other: Self) -> Self:
        Class = type(self)
        if not isinstance(other, Class):
            return NotImplemented

        return Class(self._number.__ror__(other._number))

    @override
    def __neg__(self) -> int:
        return -self._number

    @override
    def __pos__(self) -> Self:
        return type(self)(self._number)

    @override
    def __invert__(self) -> Self:
        # TODO: How should this behave?
        raise NotImplementedError()

    @override
    def __floor__(self) -> Self:
        return type(self)(self)

    @override
    def __ceil__(self) -> Self:
        return type(self)(self)

    @override
    def __int__(self) -> int:
        return self._number

    @override
    def __eq__(self, other: object) -> bool:
        # Unlike the other comparison dunder methods (eg. `__lt__`, `__ge__`,
        # etc.), `__eq__` is expected to work with any object, so mypy doesn't
        # detect comparisons between `Uint` and `int` as errors. Instead of
        # throwing a `TypeError` at runtime, we try to behave sanely and
        # soundly by converting `other` to an integer if possible, then
        # comparing.
        if isinstance(other, Unsigned):
            return self._number == other._number
        elif isinstance(other, SupportsInt):
            other_int = int(other)
            if other != other_int:
                # If `other` doesn't equal `int(other)`, `self` definitely
                # doesn't equal `other` since `self` has to be an integer.
                return False
            return self._number == other_int
        return NotImplemented

    @override
    def __le__(self, other: Self) -> bool:
        Class = type(self)
        if not isinstance(other, Class):
            return NotImplemented

        return self._number <= other._number

    def __ge__(self, other: Self) -> bool:
        Class = type(self)
        if not isinstance(other, Class):
            return NotImplemented

        return self._number >= other._number

    @override
    def __lt__(self, other: Self) -> bool:
        Class = type(self)
        if not isinstance(other, Class):
            return NotImplemented

        return self._number < other._number

    def __gt__(self, other: Self) -> bool:
        Class = type(self)
        if not isinstance(other, Class):
            return NotImplemented

        return self._number > other._number

    @override
    def __round__(self, ndigits: Optional[int] = None) -> Self:
        return type(self)(self)

    @override
    def __trunc__(self) -> Self:
        return type(self)(self)

    @override
    def __rshift__(self, right: Self) -> Self:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented

        return Class(self._number >> right._number)

    @override
    def __rrshift__(self, left: Self) -> Self:
        Class = type(self)
        if not isinstance(left, Class):
            return NotImplemented

        return Class(self._number.__rrshift__(left._number))

    @override
    def __lshift__(self, right: Self) -> Self:
        Class = type(self)
        if not isinstance(right, Class):
            return NotImplemented

        return Class(self._number << right._number)

    @override
    def __rlshift__(self, left: Self) -> Self:
        Class = type(self)
        if not isinstance(left, Class):
            return NotImplemented

        return Class(self._number.__rlshift__(left._number))

    @override
    def __hash__(self) -> int:
        return hash((type(self), self._number))

    @override
    def __repr__(self) -> str:
        return "{}({})".format(type(self).__name__, self._number)

    @override
    def __str__(self) -> str:
        return str(self._number)

    def to_be_bytes64(self) -> Bytes64:
        """
        Converts this unsigned integer into its big endian representation with
        exactly 64 bytes.
        """
        return Bytes64(self._number.to_bytes(64, "big"))

    def to_be_bytes32(self) -> Bytes32:
        """
        Converts this unsigned integer into its big endian representation
        with exactly 32 bytes.
        """
        return Bytes32(self._number.to_bytes(32, "big"))

    def to_le_bytes4(self) -> "Bytes4":
        """
        Converts this unsigned integer into its little endian representation,
        with exactly 4 bytes.
        """
        return Bytes4(self._number.to_bytes(4, "little"))

    def to_be_bytes4(self) -> "Bytes4":
        """
        Converts this unsigned integer into its big endian representation, with
        exactly 4 bytes.
        """
        return Bytes4(self._number.to_bytes(4, "big"))

    def to_le_bytes8(self) -> "Bytes8":
        """
        Converts this fixed sized unsigned integer into its little endian
        representation, with exactly 8 bytes.
        """
        return Bytes8(self._number.to_bytes(8, "little"))

    def to_be_bytes8(self) -> "Bytes8":
        """
        Converts this unsigned integer into its big endian representation, with
        exactly 8 bytes.
        """
        return Bytes8(self._number.to_bytes(8, "big"))

    def to_bytes(
        self,
        length: Optional[Self] = None,
        byteorder: Literal["big", "little"] = "big",
    ) -> Bytes:
        """
        Return an array of bytes representing an integer.
        """
        if length is None:
            length_int = 1
        else:
            length_int = int(length)
        return self._number.to_bytes(length=length_int, byteorder=byteorder)

    def to_be_bytes(self) -> "Bytes":
        """
        Converts this unsigned integer into its big endian representation,
        without padding.
        """
        bit_length = self._number.bit_length()
        byte_length = (bit_length + 7) // 8
        return self._number.to_bytes(byte_length, "big")

    def to_le_bytes(self) -> "Bytes":
        """
        Converts this unsigned integer into its little endian representation,
        without padding.
        """
        bit_length = self._number.bit_length()
        number_bytes = (bit_length + 7) // 8
        return self._number.to_bytes(number_bytes, "little")

    def to_le_bytes32(self) -> Bytes32:
        """
        Converts this unsigned integer into its little endian representation
        with exactly 32 bytes.
        """
        return Bytes32(self._number.to_bytes(32, "little"))

    def to_le_bytes64(self) -> Bytes64:
        """
        Converts this unsigned integer into its little endian representation
        with exactly 64 bytes.
        """
        return Bytes64(self._number.to_bytes(64, "little"))

    def bit_length(self) -> "Uint":
        """
        Minimum number of bits required to represent this number in binary.
        """
        return Uint(self._number.bit_length())


class Uint(Unsigned):
    """
    Unsigned integer of arbitrary size.
    """

    @classmethod
    def from_be_bytes(cls: Type[Self], buffer: Bytes) -> Self:
        """
        Converts a sequence of bytes into an arbitrarily sized unsigned integer
        from its big endian representation.
        """
        return cls(int.from_bytes(buffer, "big"))

    @classmethod
    def from_le_bytes(cls: Type[Self], buffer: Bytes) -> Self:
        """
        Converts a sequence of bytes into an arbitrarily sized unsigned integer
        from its little endian representation.
        """
        return cls(int.from_bytes(buffer, "little"))

    @override
    def _in_range(self, value: int) -> bool:
        return value >= 0


def ulen(input: Sized, /) -> Uint:
    """
    Return the number of items in a container, as a `Uint`.
    """
    return Uint(len(input))


class FixedUnsigned(Unsigned):
    """
    Superclass for fixed size unsigned integers. Not intended to be used
    directly, but rather to be subclassed.
    """

    MAX_VALUE: ClassVar[Self]
    """
    Largest value that can be represented by this integer type.
    """

    @classmethod
    def from_be_bytes(cls: Type[Self], buffer: "Bytes") -> Self:
        """
        Converts a sequence of bytes into a fixed sized unsigned integer
        from its big endian representation.
        """
        bits = cls.MAX_VALUE._number.bit_length()
        byte_count = (bits + 7) // 8
        if len(buffer) > byte_count:
            raise ValueError()

        return cls(int.from_bytes(buffer, "big"))

    @classmethod
    def from_le_bytes(cls: Type[Self], buffer: "Bytes") -> Self:
        """
        Converts a sequence of bytes into a fixed sized unsigned integer
        from its little endian representation.
        """
        bits = cls.MAX_VALUE._number.bit_length()
        byte_count = (bits + 7) // 8
        if len(buffer) > byte_count:
            raise ValueError()

        return cls(int.from_bytes(buffer, "little"))

    @classmethod
    def from_signed(cls: Type[Self], value: int) -> Self:
        """
        Creates an unsigned integer representing `value` using two's
        complement.
        """
        if value >= (cls.MAX_VALUE._number // 2 + 1):
            raise OverflowError

        if value >= 0:
            return cls(value)

        if value < (-cls.MAX_VALUE // 2):
            raise OverflowError

        return cls(value & cls.MAX_VALUE._number)

    @override
    def _in_range(self, value: int) -> bool:
        return value >= 0 and value <= self.MAX_VALUE._number

    def wrapping_add(self, right: Self) -> Self:
        """
        Return a new instance containing `self + right (mod N)`.
        """
        Class = type(self)
        if not isinstance(right, Class):
            raise TypeError()

        # This is a fast way of ensuring that the result is < (2 ** 256)
        return Class((self._number + right._number) & self.MAX_VALUE._number)

    def wrapping_sub(self, right: Self) -> Self:
        """
        Return a new instance containing `self - right (mod N)`.
        """
        Class = type(self)
        if not isinstance(right, Class):
            raise TypeError()

        # This is a fast way of ensuring that the result is < (2 ** 256)
        return Class((self._number - right._number) & self.MAX_VALUE._number)

    def wrapping_mul(self, right: Self) -> Self:
        """
        Return a new instance containing `self * right (mod N)`.
        """
        Class = type(self)
        if not isinstance(right, Class):
            raise TypeError

        # This is a fast way of ensuring that the result is < (2 ** 256)
        return Class((self._number * right._number) & self.MAX_VALUE._number)

    def wrapping_pow(self, right: Self, modulo: Optional[Self] = None) -> Self:
        """
        Return a new instance containing `self ** right (mod modulo)`.

        If omitted, `modulo` defaults to `Uint(self.MAX_VALUE) + 1`.
        """
        Class = type(self)
        modulo_int = None
        if modulo is not None:
            if not isinstance(modulo, Class):
                raise TypeError()
            modulo_int = modulo._number

        if not isinstance(right, Class):
            raise TypeError()

        # This is a fast way of ensuring that the result is < (2 ** 256)
        return Class(
            self._number.__pow__(right._number, modulo_int)
            & self.MAX_VALUE._number
        )

    @override
    def __invert__(self: Self) -> Self:
        return type(self)(
            int.__invert__(self._number) & self.MAX_VALUE._number
        )

    def to_signed(self) -> int:
        """
        Decodes a signed integer from its two's complement representation.
        """
        bits = self.MAX_VALUE._number.bit_length()
        bits = 8 * ((bits + 7) // 8)
        if self._number.bit_length() < bits:
            # This means that the sign bit is 0
            return int(self)

        # -1 * (2's complement of value)
        return int(self) - (self.MAX_VALUE._number + 1)


_V = TypeVar("_V", bound=FixedUnsigned)


def _max_value(class_: Type[_V], bits: int) -> _V:
    value = object.__new__(class_)
    value._number = (2**bits) - 1  # type: ignore[misc]
    return value


class U256(FixedUnsigned):
    """
    Unsigned integer, which can represent `0` to `2 ** 256 - 1`, inclusive.
    """

    MAX_VALUE: ClassVar["U256"]
    """
    Largest value that can be represented by this integer type.
    """


U256.MAX_VALUE = _max_value(U256, 256)


class U32(FixedUnsigned):
    """
    Unsigned positive integer, which can represent `0` to `2 ** 32 - 1`,
    inclusive.
    """

    MAX_VALUE: ClassVar["U32"]
    """
    Largest value that can be represented by this integer type.
    """


U32.MAX_VALUE = _max_value(U32, 32)


class U64(FixedUnsigned):
    """
    Unsigned positive integer, which can represent `0` to `2 ** 64 - 1`,
    inclusive.
    """

    MAX_VALUE: ClassVar["U64"]
    """
    Largest value that can be represented by this integer type.
    """


U64.MAX_VALUE = _max_value(U64, 64)



import pytest

from ethereum_types.bytes import FixedBytes


def test_fixed_bytes_init_too_short() -> None:
    class TestBytes(FixedBytes):
        LENGTH = 5

    with pytest.raises(ValueError):
        TestBytes(b"\0")


def test_fixed_bytes_init_too_long() -> None:
    class TestBytes(FixedBytes):
        LENGTH = 5

    with pytest.raises(ValueError):
        TestBytes(b"\0" * 6)


def test_fixed_bytes_init() -> None:
    class TestBytes(FixedBytes):
        LENGTH = 5

    tb = TestBytes(b"\0" * 5)
    assert tb == b"\0\0\0\0\0"


def test_fixed_bytes_init_bytearray() -> None:
    class TestBytes(FixedBytes):
        LENGTH = 5

    tb = TestBytes(bytearray([0, 0, 0, 0, 0]))
    assert tb == b"\0\0\0\0\0"
    assert isinstance(tb, bytes)
    assert not isinstance(tb, bytearray)  # type: ignore[unreachable]


def test_fixed_bytes_concat() -> None:
    class TestBytes(FixedBytes):
        LENGTH = 5

    tb0 = TestBytes(b"\0" * 5)
    tb1 = TestBytes(b"1" * 5)

    tb = tb0 + tb1

    assert tb == b"\x00\x00\x00\x00\x0011111"
    assert isinstance(tb, bytes)
    assert not isinstance(tb, TestBytes)



from dataclasses import dataclass

import pytest

from ethereum_types.frozen import modify, slotted_freezable


@slotted_freezable
@dataclass
class Mock:
    a: int
    b: str


def test_slotted_freezable_init() -> None:
    positional = Mock(1, "2")
    assert positional.a == 1
    assert positional.b == "2"

    keyword = Mock(b="2", a=1)
    assert keyword.a == 1
    assert keyword.b == "2"


def test_slotted_freezable_frozen() -> None:
    value = Mock(1, "2")
    with pytest.raises(AttributeError):
        value.a = 2

    with pytest.raises(AttributeError):
        value.b = "2"

    with pytest.raises(AttributeError):
        del value.a


def test_slotted_freezable_init_thawed_frozen() -> None:
    value = Mock(1, "2", _frozen=False)  # type: ignore[call-arg]
    value.a = 2
    value.b = "2"
    del value.a


def test_slotted_freezable_modify() -> None:
    before = Mock(1, "2")

    def f(x: Mock) -> None:
        x.a = 3

    after = modify(before, f)

    assert before.a == 1
    assert after.a == 3
    assert before.b == after.b
    assert after is not before



"""
Sequences of 256-bit values.
"""

from typing import Any, ClassVar, Type, TypeVar

B = TypeVar("B", bound="FixedBytes")


class FixedBytes(bytes):
    """
    Superclass for fixed sized byte arrays. Not intended to be used directly,
    but should be subclassed.
    """

    LENGTH: ClassVar[int]
    """
    Number of bytes in each instance of this class.
    """

    __slots__ = ()

    def __new__(cls: Type[B], *args: Any, **kwargs: Any) -> B:
        """
        Create a new instance, ensuring the result has the correct length.
        """
        result = super(FixedBytes, cls).__new__(cls, *args, **kwargs)
        if len(result) != cls.LENGTH:
            raise ValueError(
                f"expected {cls.LENGTH} bytes but got {len(result)}"
            )
        return result


class Bytes0(FixedBytes):
    """
    Byte array of exactly zero elements.
    """

    LENGTH = 0
    """
    Number of bytes in each instance of this class.
    """


class Bytes4(FixedBytes):
    """
    Byte array of exactly four elements.
    """

    LENGTH = 4
    """
    Number of bytes in each instance of this class.
    """


class Bytes8(FixedBytes):
    """
    Byte array of exactly eight elements.
    """

    LENGTH = 8
    """
    Number of bytes in each instance of this class.
    """


class Bytes20(FixedBytes):
    """
    Byte array of exactly 20 elements.
    """

    LENGTH = 20
    """
    Number of bytes in each instance of this class.
    """


class Bytes32(FixedBytes):
    """
    Byte array of exactly 32 elements.
    """

    LENGTH = 32
    """
    Number of bytes in each instance of this class.
    """


class Bytes48(FixedBytes):
    """
    Byte array of exactly 48 elements.
    """

    LENGTH = 48


class Bytes64(FixedBytes):
    """
    Byte array of exactly 64 elements.
    """

    LENGTH = 64
    """
    Number of bytes in each instance of this class.
    """


class Bytes256(FixedBytes):
    """
    Byte array of exactly 256 elements.
    """

    LENGTH = 256
    """
    Number of bytes in each instance of this class.
    """


Bytes = bytes
"""
Sequence of bytes (octets) of arbitrary length.
"""



"""
Dataclass extension that supports immutability.
"""

from dataclasses import is_dataclass, replace
from typing import (
    Any,
    Callable,
    Concatenate,
    ParamSpec,
    Protocol,
    TypeVar,
    runtime_checkable,
)


@runtime_checkable
class SlottedFreezable(Protocol):
    """
    A [`Protocol`] implemented by data classes annotated with
    [`@slotted_freezable`].

    [`@slotted_freezable`]: ref:ethereum_types.frozen.slotted_freezable
    [`Protocol`]: https://docs.python.org/library/typing.html#typing.Protocol
    """

    _frozen: bool


def _setattr_function(self: Any, attr: str, value: Any) -> None:
    if getattr(self, "_frozen", None):
        raise AttributeError("Mutating frozen dataclasses is not allowed.")
    else:
        object.__setattr__(self, attr, value)


def _delattr_function(self: Any, attr: str) -> None:
    if self._frozen:
        raise AttributeError("Mutating frozen dataclasses is not allowed.")
    else:
        object.__delattr__(self, attr)


_S = TypeVar("_S", bound=SlottedFreezable)
_P = ParamSpec("_P")


def _make_init_function(
    f: Callable[Concatenate[_S, _P], None]
) -> Callable[Concatenate[_S, _P], None]:
    def init_function(self: _S, *args: _P.args, **kwargs: _P.kwargs) -> None:
        will_be_frozen = kwargs.pop("_frozen", True)
        assert isinstance(will_be_frozen, bool)
        object.__setattr__(self, "_frozen", False)
        f(self, *args, **kwargs)
        self._frozen = will_be_frozen

    return init_function


def slotted_freezable(cls: Any) -> Any:
    """
    Monkey patches a dataclass so it can be frozen by setting `_frozen` to
    `True` and uses `__slots__` for efficiency.

    Instances will be created frozen by default unless you pass `_frozen=False`
    to `__init__`.
    """
    cls.__slots__ = ("_frozen",) + tuple(cls.__annotations__)
    cls.__init__ = _make_init_function(cls.__init__)
    cls.__setattr__ = _setattr_function
    cls.__delattr__ = _delattr_function
    return type(cls)(cls.__name__, cls.__bases__, dict(cls.__dict__))


S = TypeVar("S")


def modify(obj: S, f: Callable[[S], None]) -> S:
    """
    Create a copy of `obj` (which must be [`@slotted_freezable`]), and modify
    it by applying `f`. The returned copy will be frozen.

    [`@slotted_freezable`]: ref:ethereum_types.frozen.slotted_freezable
    """
    assert is_dataclass(obj)
    assert isinstance(obj, SlottedFreezable)
    new_obj = replace(obj, _frozen=False)  # type: ignore[unreachable]
    f(new_obj)
    new_obj._frozen = True
    return new_obj



"""
Integer and array types which are used by—but not unique to—Ethereum.

[`Uint`] represents non-negative integers of arbitrary size, while subclasses
of [`FixedUnsigned`] (like [`U256`] or [`U32`]) represent non-negative integers
of particular sizes.

Similarly, [`Bytes`] represents arbitrarily long byte sequences, while
subclasses of [`FixedBytes`] (like [`Bytes0`] or [`Bytes64`]) represent
sequences containing an exact number of bytes.

[`Uint`]: ref:ethereum_types.numeric.Uint
[`FixedUnsigned`]: ref:ethereum_types.numeric.FixedUnsigned
[`U32`]: ref:ethereum_types.numeric.U32
[`U256`]: ref:ethereum_types.numeric.U256
[`Bytes`]: ref:ethereum_types.bytes.Bytes
[`FixedBytes`]: ref:ethereum_types.bytes.FixedBytes
[`Bytes0`]: ref:ethereum_types.bytes.Bytes0
[`Bytes64`]: ref:ethereum_types.bytes.Bytes64
"""

__version__ = "0.1.0"
